// -*- c++ -*-

// This software module was originally developed by Karl Anders ygard
// (Telenor R&D) in the course of development of the MPEG-4 Systems
// standard (ISO/IEC 14496-1).  This software module is an implementation
// of a part of one or more MPEG-4 Systems tools as specified by the MPEG-4
// Systems standard.  ISO/IEC gives users of the MPEG-4 Systems standard
// free license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// For more information or to receive an up to date version of this module,
// mail karlo@nta.no or check <http://televr.fou.telenor.no/compositor/>.
//
// Copyright 1997, 1998.

#include "OpenGl/OpenGl.h"
#include "IndexedRegionSetProxy.h"  ////////
#include "opengl/base/OpenGLVisualRenderer.h"
#include "../../IM1Decoders/AFX/jwafxstreamdecoder.h"
#include "Stream.h"
#include "opengl/appearance/MultiTextureProxy.h"

// Archivos con los mtodos para ajustar las texturas y las transformaciones segn cada regin
#include "../appearance/TextureNodeProxy.h"
#include "../appearance/TextureTransformProxy.h"

// Constructor
IndexedRegionSetProxy::IndexedRegionSetProxy(IndexedRegionSet *pIndexedRegionSet): 
AfxStreamProxy(pIndexedRegionSet),
m_nTimeStamp(-1),
m_pAutoNormals(NULL),
cases_found(false)
{}


// traverse scene graph, checking for intersection
/* virtual */ BOOL
IndexedRegionSetProxy::Select(  ///////////
							OpenGLVisualRenderer* pRenderer, 
							DWORD dwTime, 
							unsigned int& depth)
{
	GLuint selectBuffer[4];

	::glSelectBuffer(4, selectBuffer);
	::glRenderMode(GL_SELECT);

	Render(pRenderer, dwTime);

//#ifdef _DEBUG
//	if (pRenderer->CheckGLError("IndexedRegionSetProxy:"))   //////////
//		TRACE("IndexedRegionSetProxy::Select() [%08x]: ", this);   ///////
//#endif

	if (::glRenderMode(GL_RENDER) > 0)
		if (depth > selectBuffer[1]) {
			depth = selectBuffer[1];

			return TRUE;
		}

		return FALSE;
}

// Render object
BOOL IndexedRegionSetProxy::RenderFromStream(MediaStream* poStream, int id, DWORD dwTime)  
{
	BYTE * pData = NULL;
	int nSize = 0;
	BOOL bEOU;
	int nStatus;
	BOOL bSkip = TRUE;
	BOOL bWait = FALSE;
	long lDelta = 0;
	BOOL bDontClose = TRUE;
	DWORD rdvTime;

	pData = (unsigned char *) poStream->Fetch(
		nSize, rdvTime, bEOU, nStatus, bSkip, bWait, lDelta, bDontClose);

	TRACE("[BitWrapperProxy::Enable] Fetch returned Size %4d \n", nSize);

	if (pData)
	{
		//ory DMLAB 2006 12
		if(type==0)
		{
			JWAFX3DMCDecoder* pDecoder = new JWAFX3DMCDecoder;

			pDecoder->SetValues(pData);

			pDecoder->Write("AFX-IndexedRegionSet.debug.vrml", id);   
		}
		else if(type==1)
		{	
			JWAFX3DMCExtensionDecoder* pDecoder = new JWAFX3DMCExtensionDecoder;

			pDecoder->SetValues(pData);

			pDecoder->Write("AFX-IndexedRegionSet.debug.vrml", id); 
		}
		//~ory DMLAB
		Enable(pData);

		return TRUE;
	}
	return FALSE;
}

BOOL IndexedRegionSetProxy::RenderFromBuffer(SFString *pBuffer, int id)  /////////
{
	BYTE * pData = NULL;
	int nSize = 0;

	//ory DMLAB 2006 12
	if(type==0)
	{
		JWAFX3DMCDecoder* pDecoder = new JWAFX3DMCDecoder;

		// Decode the bitstream
		pDecoder->Decode((LPBYTE) pBuffer->GetValue(), pBuffer->GetLength(), (LPBYTE) pData, nSize);

		pDecoder->Write("AFX-IndexedRegionSet.debug.vrml", id);  /////////
	}
	else if(type==1)
	{
		JWAFX3DMCExtensionDecoder* pDecoder = new JWAFX3DMCExtensionDecoder;

		// Decode the bitstream
		pDecoder->Decode((LPBYTE) pBuffer->GetValue(), pBuffer->GetLength(), (LPBYTE) pData, nSize);

		pDecoder->Write("AFX-IndexedRegionSet.debug.vrml", id);  ///////////
	}
	//~ory DMLAB
	Enable(pData);

	return TRUE;
}

BOOL IndexedRegionSetProxy::Render(DWORD dwTime) ////////
{
	return TRUE;
}



///// ESTE MTODO EST CAMBIADO COMPLETAMENTE
///// E INCORPORA EL inlinedRenderLoop

/* virtual */ BOOL
IndexedRegionSetProxy::Render(OpenGLVisualRenderer* pRenderer, DWORD dwTime) 
{
	IndexedRegionSet* pIndexedRegionSet = (IndexedRegionSet*) GetMediaObject();
	ComputeNormals();
	// Encontrar las variables *_case si no han sido encontradas antes
	if (!cases_found) {
		//     N : 0
		//    PO_V : 3    PI_V : 4   RPI_V : 5                                         RPI_F : 9
		//  RO_V : 6              RI_V : 8                      RO_F : 10           RI_F : 12
		// ARTIF : 13
		cases_found = true;

		int color_case_int = 0;
		if (GetColor())
			color_case_int = 3;
		else if (GetColor(0))
			color_case_int = 6;
		if (GetColorIndex().GetLength() != 0)
			color_case_int += 1;
		else if (GetColorIndex(0).GetLength() != 0)
			color_case_int += 2;
		if (color_case_int > 4)
			color_case_int = ((BOOL) (Bool) pIndexedRegionSet->colorPerVertex) ? 0 : 4;
		color_this_vtx = false;
		color_this_face = false;
		if (color_case_int>8)
			color_this_face = true;
		else if (color_case_int != 0)
			color_this_vtx = true;
		color_case = (Attrib_case) color_case_int;

		int normal_case_int = 0;
		if (GetNormal())
			normal_case_int = 3;
		else if (GetNormal(0))
			normal_case_int = 6;
		if (GetNormalIndex().GetLength() != 0)
			normal_case_int += 1;
		else if (GetNormalIndex(0).GetLength() != 0)
			normal_case_int += 2;
		if (normal_case_int > 4)
			normal_case_int = ((BOOL) (Bool)pIndexedRegionSet->normalPerVertex) ? 0 : 4;
		if (normal_case_int == 0)
			normal_case_int = 13; 
		normal_this_vtx = false;
		normal_this_face = false;
		// Aqu tengo en cuenta que "normal_case" nunca puede valer 0, por lo que siempre
		// normal_this_vtx o normal_this_face va a ser true
		if ((normal_case_int>8) && (normal_case_int<13))
			normal_this_face = true;
		else
			normal_this_vtx = true;
		normal_case = (Attrib_case) normal_case_int;
		
		int tex_case_int = 0;
		if (GetTexCoord())
			tex_case_int = 3;
		else if (GetTexCoord(0))
			tex_case_int = 6;
		if (GetTexCoordIndex().GetLength() != 0)
			tex_case_int += 1;
		else if (GetTexCoordIndex(0).GetLength() != 0)
			tex_case_int += 2;
		tex_this_vtx = tex_case != 0;
		tex_case = (Attrib_case) tex_case_int;
	}

	// Calcular los tiempos de la ltima actualizacin. La decisin va en la variable booleana newrender
	bool newrender;
	{
		Coordinate * pCoord = GetCoord();
		Color * pColor = GetColor();
		Normal * pNormal = GetNormal();
		Normal * pNormal2 = GetAutoNormal();
		TextureCoordinate * pTexCoord = GetTexCoord();
		long max_t = pIndexedRegionSet->GetUpdateTime();
		if ((pCoord)&&(long(pCoord->GetUpdateTime())>max_t)) max_t = long(pCoord->GetUpdateTime());
		if ((pColor)&&(long(pColor->GetUpdateTime())>max_t)) max_t = long(pColor->GetUpdateTime());
		if ((pNormal)&&(long(pNormal->GetUpdateTime())>max_t)) max_t = long(pNormal->GetUpdateTime());
		if ((pNormal2)&&(long(pNormal2->GetUpdateTime())>max_t)) max_t = long(pNormal2->GetUpdateTime());
		if ((pTexCoord)&&(long(pTexCoord->GetUpdateTime())>max_t)) max_t = long(pTexCoord->GetUpdateTime());
		for (int i = 0; i < pIndexedRegionSet->regions.GetLength(); i++) {
			if (long(GetRegion(i)->GetUpdateTime())>max_t) max_t = long(GetRegion(i)->GetUpdateTime());
			pColor = GetColor(i);
			pNormal = GetNormal(i);
			pTexCoord = GetTexCoord(i);
			if ((pColor)&&(long(pColor->GetUpdateTime())>max_t)) max_t = long(pColor->GetUpdateTime());
			if ((pNormal)&&(long(pNormal->GetUpdateTime())>max_t)) max_t = long(pNormal->GetUpdateTime());
			if ((pTexCoord)&&(long(pTexCoord->GetUpdateTime())>max_t)) max_t = long(pTexCoord->GetUpdateTime());
		}
		newrender = max_t > m_nTimeStamp;
	}

	if (newrender) {

		// Tener en cuenta solid, ccw, ...
		::glPushAttrib(GL_POLYGON_BIT | GL_CURRENT_BIT); 
		if ((BOOL) (Bool) (pIndexedRegionSet->solid))
			::glEnable(GL_CULL_FACE);
		if (! (BOOL) (Bool) (pIndexedRegionSet->ccw) )
			::glFrontFace(GL_CW);
		if (!( GetNormal()||GetNormal(0) ))
			::glDisable(GL_LIGHTING);


		if (!m_nDisplayList)								
			m_nDisplayList = ::glGenLists(1);						
		m_nTimeStamp  = (long) dwTime;
		::glNewList(m_nDisplayList, GL_COMPILE_AND_EXECUTE);

		// Crear una lista de todos los arrays que se vayan a usar
		Coordinate * pCoord = pIndexedRegionSet->coord;
		const MF<Vec3f>& coordarray = pCoord->point;
		MF<int> coordIndex; // Esto antes era MF<int>& coordIndex; pero el compilador
		// daba error porque deca que las referencias tenan que ser inicializadas en
		// la declaracin. Tambin he quitado el "&" en los tres ndices declarados aqu abajo,
		// as como en la declaracin de las variables que acaban en "array"

		Color * pColor;
		MF<Color3f> colorarray;
		MF<int> colorIndex;

		Normal * pNormal;
		MF<Vec3f> normalarray;
		MF<int> normalIndex;

		TextureCoordinate * pTexCoord;
		MF<Vec2f> texCoordarray;
		MF<int> texCoordIndex;

		// Inicializar los arrays que son iguales para todas las regiones
		// Depende del tipo de especificacin para cada atributo
		// "coord" ya se ha inicializado antes
		switch (color_case) {
			case PI_V:
				colorIndex = pIndexedRegionSet->colorIndex;
			case PO_V:
			case RPI_V:
			case RPI_F:
				pColor = pIndexedRegionSet->color;
				colorarray = pColor->color;
		}

		switch (normal_case) {
			case PI_V:
				normalIndex = pIndexedRegionSet->normalIndex;
			case PO_V:
			case RPI_V:
			case RPI_F:
				pNormal = GetNormal();
				normalarray = pNormal->vector;
				break;
			case ARTIF:					
				pNormal = GetAutoNormal();
				normalarray = pNormal->vector;
		}

		switch (tex_case) {
			case PI_V:
				texCoordIndex = pIndexedRegionSet->texCoordIndex;
			case PO_V:
			case RPI_V:
				pTexCoord = GetTexCoord();
				texCoordarray = pTexCoord->point;
		}

		// Crear position_reorder si hay algn atributo que lo vaya a utilizar
		// Es decir, si alguna es RPI_V, RO_V o RI_V
		// No quiere decir que se rellene aqu, slo que se reserva el sitio
		bool reorder = false;
		int* position_reorder;
		int* last_known_region;
		if (( color_case == RPI_V) || ( color_case == RO_V) || ( color_case == RI_V) || 
			(normal_case == RPI_V) || (normal_case == RO_V) || (normal_case == RI_V) || 
			(   tex_case == RPI_V) || (   tex_case == RO_V) || (   tex_case == RI_V)) {
				position_reorder = new int[coordarray.GetLength()];
				last_known_region = new int[coordarray.GetLength()];
				for (int l=0; l<coordarray.GetLength(); l++)
					last_known_region[l] = -1;
				reorder = true;
		}

		// En caso de que las normales sean artificiales ...
		int artif_normal_count = 0;

		// Recorremos cada regin
		for (int i = 0; i < pIndexedRegionSet->regions.GetLength(); i++) {
			int face = 0;


			// Inicializar los arrays que son distintos para cada regin
			// Esto depende de la forma en la que estn especificados los atributos
			// El primero es coordIndex, que siempre va por regiones
			coordIndex = GetCoordIndex(i);
			// Luego van los de los atributos
			switch (color_case) {
				case RI_V:
				case RI_F:
					colorIndex = GetColorIndex(i);
				case RO_V:
				case RO_F:
					pColor = GetColor(i);
					colorarray = pColor->color;
					break;
				case RPI_V:
				case RPI_F:
					colorIndex = GetColorIndex(i);
			}
			switch (normal_case) {
				case RI_V:
				case RI_F:
					normalIndex = GetNormalIndex(i);
				case RO_V:
				case RO_F:
					pNormal = GetNormal(i);
					normalarray = pNormal->vector;
					break;
				case RPI_V:
				case RPI_F:
					normalIndex = GetNormalIndex(i);
			}
			switch (tex_case) {
				case RI_V:
					texCoordIndex = GetTexCoordIndex(i);
				case RO_V:
					pTexCoord = GetTexCoord(i);
					texCoordarray = pTexCoord->point;
					break;
				case RPI_V:
					texCoordIndex = GetTexCoordIndex(i);
			}

			// Mirar si hay que corregir las texturas y las transformaciones de la textura
			if (GetRegion(i)->texture)
				((TextureNodeProxy*) (GetRegion(i)->texture->GetRenderingObject()))->Enable(pRenderer, dwTime);
			if (GetRegion(i)->textureTransform)
				((TextureTransformProxy*) (GetRegion(i)->textureTransform->GetRenderingObject()))->Enable(dwTime);



			// Reordenar los vrtices si hay algn modo que lo necesita
			if (reorder) {
				int vtx_in_this_region = 0;
				for (int j=0; j<coordIndex.GetLength(); j++){
					if (last_known_region[coordIndex[j]] != i) {
						last_known_region[coordIndex[j]] = i;
						vtx_in_this_region++;
					}
				}
				int* vtx_in_order = new int [vtx_in_this_region];
				int count = 0;
				for (j=0; j<coordarray.GetLength(); j++) {
					if (last_known_region[j] == i)
						vtx_in_order[count++] = j;
				}						
				// Actualizar position_reorder
				for (j=0; j<vtx_in_this_region; j++)
					position_reorder[vtx_in_order[j]] = j;
			}
			BOOL begin = false;
			int coords = coordIndex.GetLength();
			Color3f col;
			Vec3f norm;
			Vec2f texCoord;
			for (int k=0; k<coords; k++) {
				int vtx = coordIndex[k];
				if (vtx == -1) {
					if (begin) {
						::glEnd();
						begin = FALSE;
						face++;
					}
				} else {
					// Lo primero es sacar las coordenadas del vrtice que estamos tratando
					// y ponerlas en "vertex"
					const Vec3f& vertex = coordarray[vtx];
					// ____ Atributos por vrtice ____ (casos PO_V, PI_V, RPI_V, RO_V, RI_V, ARTIF)
					// Colores
				
					switch (color_case) {
						case PO_V:
							col = colorarray[vtx];
							break;
						case PI_V:
							col = colorarray[colorIndex[vtx]];
							break;
						case RPI_V:
						case RI_V:
							col = colorarray[colorIndex[position_reorder[vtx]]];
							break;
						case RO_V:
							col = colorarray[position_reorder[vtx]];
							break;
					}		
					switch (normal_case) {
						case PO_V:
							norm = normalarray[vtx];
							break;
						case PI_V:
							norm = normalarray[normalIndex[vtx]];
							break;
						case RPI_V:
						case RI_V:
							norm = normalarray[normalIndex[position_reorder[vtx]]];
							break;
						case RO_V:
							norm = normalarray[position_reorder[vtx]];
							break;
						case ARTIF:
							norm = normalarray[artif_normal_count++];
							break;
					}			
					switch(tex_case) {
						case PO_V:
							texCoord = texCoordarray[vtx];
							break;
						case PI_V:
							texCoord = texCoordarray[texCoordIndex[vtx]];
							break;
						case RPI_V:
						case RI_V:
							texCoord = texCoordarray[texCoordIndex[position_reorder[vtx]]];
							break;
						case RO_V:
							texCoord = texCoordarray[position_reorder[vtx]];
							break;
					}
					// Aplicamos los colores, normales y texturas en lenguaje OpenGL
					if(color_this_vtx)
						glColor3f(col.r, col.g, col.b);
					if (normal_this_vtx)
						glNormal3f(norm.x, norm.y, norm.z);
					if (tex_this_vtx)
						glTexCoord2f(texCoord.x, texCoord.y);


					if (!begin) {
						::glBegin(GL_POLYGON);
						// Atributos especificados por cara
						switch (color_case) {
							case RPI_F:
							case RI_F:
								col = colorarray[colorIndex[face]];
								break;
							case RO_F:
								col = colorarray[face];
								break;
						}
						switch (normal_case) {
							case RPI_F:
							case RI_F:
								norm = normalarray[normalIndex[face]];
								break;
							case RO_F:
								norm = normalarray[face];
								break;
						}
						if (color_this_face)
							glColor3f(col.r, col.g, col.b);
						if (normal_this_face)
							glNormal3f(norm.x, norm.y, norm.z);


						begin = TRUE;
					}
					glVertex3f(vertex.x, vertex.y, vertex.z);
				}
			}
			if (begin)
				::glEnd();

			// Volver a restaurar las texturas y las transformaciones
			if (GetRegion(i)->texture)
				((TextureNodeProxy*) (GetRegion(i)->texture->GetRenderingObject()))->Disable(pRenderer, dwTime);
			if (GetRegion(i)->textureTransform)
				((TextureTransformProxy*) (GetRegion(i)->textureTransform->GetRenderingObject()))->Disable(dwTime);



		}
		::glEndList();
		if (reorder) {
			delete [] position_reorder;
			delete [] last_known_region;
		}

		// Deshacer lo de "Tener en cuenta solid, ccw, ..."
		// ste es el compaero de ::glPushAttrib(GL_POLYGON_BIT | GL_CURRENT_BIT);
		::glPopAttrib();

	} else {
		::glCallList(m_nDisplayList);
	}
	return TRUE;
}

// Render object description
// Para qu sirve este mtodo?

/* virtual */ void
IndexedRegionSetProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD /* dwTime */)
{
	//int children = ((GetCoord() ? 1 : 0) +
	//	(GetColor() ? 1 : 0) +
	//	(GetNormal() ? 1 : 0) +
	//	(GetTexCoord() ? 1 : 0));

	//GLfloat color[] = { 1.0f, 0.0f, 1.0f, 0.75f};

	////satish changed this line for compilation purposes
	////have to be fixed afterwards

	////::glColor4fv(color);
	//glColor4fv(color);

	//RenderDescriptionBox(pRenderer, "IndexedRegionSet");

	//::glPushMatrix();

	//::glScalef(1.0f / children, 1.0f, 1.0f);
	//::glTranslatef(0.0f, -1.0f, 0.0f);

	//// Hay que implementar todos los mtodos especificados aqu

	//if (GetCoord()) {
	//	RenderDescriptionBox(pRenderer, "Coordinate");

	//	::glTranslatef(1.0f, 0.0f, 0.0f);
	//}

	//if (GetColor()) {
	//	RenderDescriptionBox(pRenderer, "Color");

	//	::glTranslatef(1.0f, 0.0f, 0.0f);
	//}

	//if (GetNormal()) {
	//	RenderDescriptionBox(pRenderer, "Normal");

	//	::glTranslatef(1.0f, 0.0f, 0.0f);
	//}

	//if (GetTexCoord())
	//	RenderDescriptionBox(pRenderer, "TextureCoordinate");

	//::glPopMatrix();
}


// Update bounding box

void IndexedRegionSetProxy::UpdateBbox()
{
	// Node::UpdateBbox();
	for (int i = m_CollisionParents.GetLength() - 1; i >= 0; --i)
		( (MediaObjectProxy*) ( m_CollisionParents.ElementAt(i) ->
		GetRenderingObject() ) ) ->
		UpdateBbox();
}


void IndexedRegionSetProxy::ComputeNormals()
{
	if (   (!GetNormal())   &&   (!GetNormal(0))   &&   (!GetAutoNormal())   ) {
		m_pAutoNormals = new Normal;
		m_pAutoNormals->SetExecutive(GetIndexedRegionSet()->GetExecutive());
		MediaObjectProxy * poNormalProxy = new MediaObjectProxy(m_pAutoNormals);
		m_pAutoNormals->SetRenderingObject(poNormalProxy);				
		GenerateNormals();
	}
}


void IndexedRegionSetProxy::GenerateNormals() {

	// autoNormals es una variable de la clase proxy en la que se almacenan las normales para cada esquina
	// Primero todas las esquinas de la primera regin, luego las de la segunda, ...
	// Es lo que se va a rellenar en este mtodo

	const MF<Vec3f>& coordarray = GetCoord()->point;
	ZArray<int> vtx_last_reg;
	ZArray<int> vtx_map;
	vtx_last_reg.Expand(coordarray.GetLength());
	vtx_map.Expand(coordarray.GetLength());
	for (int i=0; i<coordarray.GetLength(); i++)
		vtx_last_reg[i] = -1;

	for (int k=0; k<GetIndexedRegionSet()->regions.GetLength(); k++) {
		// Paso 1: crear una normal para cada cara con sus tres primeros vrtices a 
		// la vez que se crea una lista de las caras a las que pertenece cada vrtice
		const MF<int>& coordIndex = GetCoordIndex(k);
		ZArray<ZArray<int> > vertexDependencies;
		ZArray<Vec3f> faceNormals;	
		int point = 0;
		int face = 0;
		int numberOfUniqueVertices = 0;
		for (int i=0; i < coordIndex.GetLength(); i++) {
			int coord = coordIndex[i];
			if (coord != -1) {
				// Primero hay que ver si es la primera vez que se procesa este vrtice en
				// esta regin. Esto pasa cuando su entrada en vtx_last_reg es menor que k
				if (vtx_last_reg[coord] < k) {
					// En ese caso, hay que actualizar vtx_last_reg ...
					vtx_last_reg[coord] = k;
					// ... y darle una posicin en vertexDependencies, que guardamos en la entrada
					// que este vrtice tiene en vtx_map
					++vertexDependencies;
					vtx_map[coord] = numberOfUniqueVertices++;
				}
				// Ahora que ya estamos seguros de que
				// est visitado, se puede poner la cara actual en la entrada de este
				// vrtice en vertexDependencies
				vertexDependencies[vtx_map[coord]] += face;
				// Calculamos la normal a la cara si es el tercer punto
				if (i - point == 2) {
					Vec3f normal = ComputeNormal(
						coordarray[coordIndex[point]],
						coordarray[coordIndex[point + 1]],
						coordarray[coordIndex[point + 2]]);
					normal.Normalize();
					faceNormals += normal;
				}
			} else {
				face++;
				point = i + 1;
			}
		}
		// Ahora ya tenemos las normales de las caras de esta regin, as como las relaciones
		// entre puntos y caras, tambin dentro de la regin. Tenemos que aadir la normal de cada
		// esquina a autoNormals, teniendo en cuenta creaseAngle. As es como se hace en IndexedFaceSet
		face = 0;
		int counter = 0;
		MF<Vec3f> normals;
		for (int j = 0; j<coordIndex.GetLength(); ++j) {
			if (coordIndex[j] != -1) {
				// normals se actualiza para cada esquina, es decir, para cada entrada en
				// coordIndex distinta de -1
				const ZArray<int>& faces = vertexDependencies[vtx_map[coordIndex[j]]];
				Vec3f normal = faceNormals[face];
				Vec3f vertexNormal;
				for (int k = 0; k < faces.GetLength(); ++k)
					if (normal == faceNormals[faces[k]] ||  ::acos(normal * faceNormals[faces[k]]) <= GetIndexedRegionSet()->creaseAngle)
						vertexNormal += faceNormals[faces[k]];
				normals += vertexNormal;
			} else {
				++face;
			}			
		}
		m_pAutoNormals->vector = (normals);
	}
	m_pAutoNormals->SetUpdateTime();
}


// Compute normal

Vec3f IndexedRegionSetProxy::ComputeNormal(
	const Vec3f& point_a,
	const Vec3f& point_b,
	const Vec3f& point_c)
{
	Vec3f vector_a(point_a);
	Vec3f vector_b(point_a);

	vector_a -= point_b;
	vector_b -= point_c;

	vector_a.Cross(vector_b);
	vector_a.Normalize();

	return vector_a;
}


void IndexedRegionSetProxy::Enable (unsigned char *pData)
{
}