// -*- c++ -*-

// This software module was originally developed by Karl Anders ygard
// (Telenor R&D) in the course of development of the MPEG-4 Systems
// standard (ISO/IEC 14496-1).  This software module is an implementation
// of a part of one or more MPEG-4 Systems tools as specified by the MPEG-4
// Systems standard.  ISO/IEC gives users of the MPEG-4 Systems standard
// free license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.	 The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.	 This copyright notice must be included in
// all copies or derivative works.
//
// For more information or to receive an up to date version of this module,
// mail karlo@nta.no or check <http://televr.fou.telenor.no/compositor/>.
//
// Copyright 1997, 1998.

#ifndef __INDEXEDREGIONSETPROXY_H__
#define __INDEXEDREGIONSETPROXY_H__

#include "..\AFX\AfxStreamProxy.h"

// class IndexedRegionSet;
// class TextureCoordinate;
//#include "Nodes.h"
//#include "NodeFld.h"

#include "MeshDecoder/mesh_decode.h" // SOLARIS PORTING - case typo

/** IndexedRegionSet node.

The IndexedRegionSetProxy class specifies a three dimensional mesh, located in
origo of the local coordinate system. */

typedef enum { 
	NO_DEF=0, 
		PO_V=3, 
		PI_V=4, 
		RPI_V=5, 
		RO_V=6, 
		RI_V=8, 
		RPI_F=9, 
		RO_F=10, 
		RI_F=12, 
		ARTIF=13
} Attrib_case;

class IndexedRegionSetProxy : public AfxStreamProxy {												

private:												
	/// Automatically generated normals											
	Normal * m_pAutoNormals;												

	/// Mesh decoder											
	MeshDecoder *  m_pMeshDecoder;											
	Matrix m_OldProjMatrix;											

	ZArray<Collision*> m_CollisionParents;

	Attrib_case color_case;
	Attrib_case normal_case;
	Attrib_case tex_case;
	bool cases_found;
	bool color_this_vtx;
	bool normal_this_vtx;
	bool tex_this_vtx;
	bool color_this_face;
	bool normal_this_face;

public:												
	// Time stamp of display list											
	long m_nTimeStamp;											

	/// Constructor											
	IndexedRegionSetProxy(IndexedRegionSet* pIndexedRegionSet);											

	/// Render object											
	virtual BOOL  Render(OpenGLVisualRenderer* pRenderer, DWORD dwTime);											

	/// from the AfxStreamProxy											
	virtual BOOL  RenderFromStream(MediaStream* poStream, int id, DWORD dwTime);											
	virtual BOOL  RenderFromBuffer(SFString *buffer, int id);											
	virtual BOOL  Render(DWORD dwTime);											

	/// Traverse scene graph, checking for intersection											
	virtual BOOL Select(OpenGLVisualRenderer* pRenderer, DWORD dwTime, unsigned int& depth);											

	/// Render object description											
	virtual void RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD dwTime);											

	/// Get pointer to related IndexedRegionSet object											
	inline IndexedRegionSet * GetIndexedRegionSet() const { return (IndexedRegionSet *) GetMediaObject(); }											

	/// Add collision parent of object  											
	inline void AddCollisionParent(Collision * pParent) { m_CollisionParents += pParent; } 											

	/// Remove collision parent of object											
	inline BOOL RemoveCollisionParent(Collision * pParent) { return m_CollisionParents -= pParent; }											

	/// Update bounding box											
	virtual void UpdateBbox();											
	/// Compute Normals											
	void ComputeNormals();											
	/// Generate Normals											
	void GenerateNormals();											
	/// Compute Normal to a triple of points											
	Vec3f ComputeNormal(const Vec3f& point_a, const Vec3f& point_b, const Vec3f& point_c);	
	// Normals
	inline Normal * GetNormal() {
		return (Normal*)(MediaObject*)GetIndexedRegionSet()->normal;
	}	
	inline Normal * GetNormal(int r) {
		return (Normal*)(MediaObject*)((Region*)GetIndexedRegionSet()->regions[r])->normal;
	}	
	inline MF<int>& GetNormalIndex() { 
		return GetIndexedRegionSet()->normalIndex; 
	}	
	inline MF<int>& GetNormalIndex(int r) {
		return ((Region*)GetIndexedRegionSet()->regions[r])->normalIndex;
	}
	// Especial para las normales
	inline Normal * GetAutoNormal() {
		return m_pAutoNormals;
	}

	// Textures
	inline TextureCoordinate * GetTexCoord() {
		return (TextureCoordinate*)(MediaObject*)GetIndexedRegionSet()->texCoord;
	}	
	inline TextureCoordinate * GetTexCoord(int r) {
		return (TextureCoordinate*)(MediaObject*)((Region*)GetIndexedRegionSet()->regions[r])->texCoord;
	}	
	inline MF<int>& GetTexCoordIndex() { 
		return GetIndexedRegionSet()->texCoordIndex; 
	}	
	inline MF<int>& GetTexCoordIndex(int r) {
		return ((Region*)GetIndexedRegionSet()->regions[r])->texCoordIndex;
	}

	// Colors
	inline Color * GetColor() {
		return (Color*)(MediaObject*)GetIndexedRegionSet()->color;
	}	
	inline Color * GetColor(int r) {
		return (Color*)(MediaObject*)((Region*)GetIndexedRegionSet()->regions[r])->color;
	}	
	inline MF<int>& GetColorIndex() { 
		return GetIndexedRegionSet()->colorIndex; 
	}	
	inline MF<int>& GetColorIndex(int r) {
		return ((Region*)GetIndexedRegionSet()->regions[r])->colorIndex;
	}

	// Coordinates
	inline Coordinate * GetCoord() {
		return GetIndexedRegionSet()->coord;
	}
	inline MF<int>& GetCoordIndex(int r) {
		return ((Region*)GetIndexedRegionSet()->regions[r])->coordIndex;
	}

	// Regions
	// Devuelve un puntero a la regin
	inline Region * GetRegion(int r) {
		return (Region*)GetIndexedRegionSet()->regions[r];
	}


#if defined SAIT_CONFORMANCE_2003												
	void Enable (unsigned char *pData);											
#endif SAIT_CONFORMANCE_2003												
};												
												
												
												
#endif // __INDEXEDREGIONSETPROXY_H__												























